# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import subprocess

from enum import Enum
from pathlib import Path
from tempfile import NamedTemporaryFile

from autohooks.utils import exec_git

__all__ = [
    'exec_git',
    'get_staged_status',
    'get_status',
    'is_partially_staged_status',
    'is_staged_status',
    'stage_files_from_status_list',
    'stash_unstaged_changes',
]


def _get_git_toplevel_path():
    try:
        git_dir = exec_git('rev-parse', '--show-toplevel').rstrip()
    except subprocess.CalledProcessError as e:
        print(
            'could not determine toplevel directory. {}'.format(
                e.output.decode()
            )
        )
        raise e
    return Path(git_dir).resolve()


class Status(Enum):
    UNMODIFIED = ' '
    MODIFIED = 'M'
    ADDED = 'A'
    DELETED = 'D'
    RENAMED = 'R'
    COPIED = 'C'
    UPDATED = 'U'
    UNTRACKED = '?'
    IGNORED = '!'


class StatusEntry:
    def __init__(self, status_string, root_path=None):
        status = status_string[:2]
        filename = status_string[3:]

        self.index = Status(status[0])
        self.working_tree = Status(status[1])
        self.root_path = root_path

        if self.index == Status.RENAMED:
            new_filename, old_filename = filename.split('\0')
            self.path = Path(new_filename)
            self.old_path = Path(old_filename)
        else:
            self.path = Path(filename)

    def __str__(self):
        return '{}{} {}'.format(
            self.index.value, self.working_tree.value, str(self.path)
        )

    def __repr__(self):
        return '<StatusEntry {}>'.format(str(self))

    def absolute_path(self):
        if self.root_path:
            return (self.root_path / self.path).resolve()
        return self.path.resolve()


def _parse_status(output):
    output = output.rstrip('\0')
    if not output:
        raise StopIteration()

    output = output.split('\0')
    while output:
        line = output.pop(0)
        if line[0] == Status.RENAMED.value:
            yield '{}\0{}'.format(line, output.pop(0))
        else:
            yield line


def is_staged_status(status):
    return (
        status.index != Status.UNMODIFIED
        and status.index != Status.UNTRACKED
        and status.index != Status.IGNORED
        and status.index != Status.DELETED
    )


def is_partially_staged_status(status):
    return (
        status.index != Status.UNMODIFIED
        and status.index != Status.UNTRACKED
        and status.index != Status.IGNORED
        and status.index != Status.DELETED
        and status.working_tree != Status.UNMODIFIED
        and status.working_tree != Status.UNTRACKED
        and status.working_tree != Status.IGNORED
    )


def get_status(files=None):
    args = [
        'status',
        '--porcelain=v1',
        '-z',
        '--ignore-submodules',
        '--untracked-files=no',
    ]

    if files is not None:
        args.append('--')
        args.extend([str(f) for f in files])

    output = exec_git(*args)
    root_path = _get_git_toplevel_path()
    return [StatusEntry(f, root_path) for f in _parse_status(output)]


def get_staged_status(files=None):
    status = get_status(files)
    return [s for s in status if is_staged_status(s)]


def stage_files_from_status_list(status_list):
    filenames = [str(s.path) for s in status_list]
    exec_git('add', *filenames)


def _write_tree():
    return exec_git('write-tree').strip()


def _read_tree(ref_or_hashid):
    exec_git('read-tree', ref_or_hashid)


def _checkout_from_index(status_list):
    filenames = [str(s.path) for s in status_list]
    exec_git('checkout-index', '-f', '--', *filenames)


def _set_ref(name, hashid):
    exec_git('update-ref', name, hashid)


def _get_tree_diff(tree1, tree2):
    return subprocess.check_output(
        [
            'git',
            'diff-tree',
            '--ignore-submodules',
            '--binary',
            '--no-color',
            '--no-ext-diff',
            '--unified=0',
            tree1,
            tree2,
        ]
    )


def _apply_diff(patch):
    with NamedTemporaryFile(mode='wb', buffering=0) as f:
        f.write(patch)

        exec_git(
            'apply',
            '-v',
            '--whitespace=nowarn',
            '--reject',
            '--recount',
            '--unidiff-zero',
            f.name,
        )


INDEX_REF = 'refs/autohooks/index'
WORKING_REF = 'refs/autohooks/working'


class stash_unstaged_changes:
    def __init__(self, status_list):
        self.partially_staged = [
            s for s in status_list if is_partially_staged_status(s)
        ]

    def stash_changes(self):
        # save current staging area aka. index
        self.index = _write_tree()
        # add ref to be able to restore index manually
        _set_ref(INDEX_REF, self.index)
        # add changes from files to index
        stage_files_from_status_list(self.partially_staged)
        # save index as working tree
        self.working_tree = _write_tree()
        # add ref to be able to restore working tee manually
        _set_ref(WORKING_REF, self.working_tree)

        # restore index without working tree changes
        # working tree changes are "stashed" now
        _read_tree(self.index)
        _checkout_from_index(self.partially_staged)

    def restore_working_tree(self):
        # restore working tree
        _read_tree(self.working_tree)
        # checkout working tree
        _checkout_from_index(self.partially_staged)

    def __enter__(self):
        if self.partially_staged:
            self.stash_changes()

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.partially_staged:
            return

        if exc_type is not None:
            # an error has occurred
            # restore working tree and index as it was before formatting
            self.restore_working_tree()
            _read_tree(self.index)
        else:
            # save formatting changes
            formatted_tree = _write_tree()

            self.restore_working_tree()

            # restore index
            # formatted_tree will be the same as index if no changes are applied
            _read_tree(formatted_tree)

            if formatted_tree != self.index:
                # create diff between index and formatted_tree
                patch = _get_tree_diff(self.index, formatted_tree)
                # apply diff to working tree
                _apply_diff(patch)



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..install import (
    install_pre_commit_hook,
    get_pre_commit_hook_path,
    get_pre_commit_hook_template_path,
)


def check_hooks():
    pre_commit_hook = get_pre_commit_hook_path()
    pre_commit_template = get_pre_commit_hook_template_path()

    template = pre_commit_template.read_text()

    if pre_commit_hook.is_file():
        hook = pre_commit_hook.read_text()
        if hook == template:
            print('autohooks pre-commit hook is active.')
        else:
            print(
                'autohooks pre-commit hook is not active. But a different '
                'pre-commit hook has been found at {}.'.format(
                    str(pre_commit_hook)
                )
            )

    else:
        print(
            'autohooks pre-commit hook not active. Please run \'autohooks '
            'activate\'.'
        )



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import sys

from autohooks.config import load_config_from_pyproject_toml


def run():
    print('autohooks => pre-commit')

    config = load_config_from_pyproject_toml()
    for name in config.get_pre_commit_script_names():
        try:
            script = importlib.import_module(name)
            retval = script.precommit()

            if retval:
                return retval

        except ImportError as e:
            print(
                'An error occurred while importing pre-commit '
                'hook {}. {}. The hook will be ignored.'.format(name, e),
                file=sys.stderr,
            )
        except Exception as e:
            print(
                'An error occurred while running pre-commit '
                'hook {}. {}. The hook will be ignored.'.format(name, e),
                file=sys.stderr,
            )
    return 0



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import toml

from .utils import get_pyproject_toml_path


class Config:
    def __init__(self, config_dict=None):
        self._config = config_dict

    def has_config(self):
        return self._config is not None

    def is_autohooks_enabled(self):
        return self.has_config()

    def get_pre_commit_script_names(self):
        if self.has_config():
            return self._config.get('pre-commit', [])

        return []


def load_config_from_pyproject_toml(pyproject_toml=None):
    if pyproject_toml is None:
        pyproject_toml = get_pyproject_toml_path()

    if not pyproject_toml.exists():
        return Config()

    config_dict = toml.load(str(pyproject_toml))
    autohooks_config = config_dict.get('tool', {}).get('autohooks')
    return Config(autohooks_config)



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from pathlib import Path


def exec_git(*args, ignore_errors=False):
    try:
        cmd_args = ['git']
        cmd_args.extend(args)
        output = subprocess.check_output(cmd_args)
        return output.decode()
    except subprocess.CalledProcessError as e:
        if ignore_errors:
            return ''
        raise e


def get_git_directory_path():
    pwd = os.environ['PWD']
    try:
        git_dir = exec_git('-C', pwd, 'rev-parse', '--git-dir').rstrip()
    except subprocess.CalledProcessError as e:
        print(
            'could not determine .git directory. {}'.format(e.output.decode())
        )
        raise e

    if pwd and not pwd in git_dir:
        git_dir_path = Path(pwd) / git_dir
    else:
        git_dir_path = Path(git_dir)

    return git_dir_path.resolve()


def get_autohooks_directory_path():
    return Path(__file__).resolve().parent


def get_git_hook_directory_path():
    git_dir_path = get_git_directory_path()
    return git_dir_path / 'hooks'


def is_project_root(path):
    return (
        (path / 'pyproject.toml').is_file()
        or (path / '.git').is_dir()
        or (path / 'setup.py').is_file()
        or (path / 'setup.cfg').is_file()
    )


def get_project_root_path():
    path = Path(os.environ['PWD'])
    path.resolve()

    if is_project_root(path):
        return path

    for parent in path.parents:
        if is_project_root(parent):
            return parent

    return path


def get_pyproject_toml_path():
    root = get_project_root_path()
    return root / 'pyproject.toml'



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse

from autohooks.version import get_version

from autohooks.cli.activate import install_hooks
from autohooks.cli.check import check_hooks

DESCRIPTION = 'autohooks - Manage git hooks'


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(get_version()),
    )

    subparsers = parser.add_subparsers(dest='command')
    activate_parser = subparsers.add_parser(
        'activate', help='Activate the pre-commit hook.'
    )
    activate_parser.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='Force activation of hook even if a hook already exists',
    )

    subparsers.add_parser('check', help='Check installed pre-commit hook')

    args = parser.parse_args()

    if not args.command:
        parser.print_usage()

    if args.command == 'activate':
        install_hooks(args)
    elif args.command == 'check':
        check_hooks()


if __name__ == "__main__":
    main()



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from pathlib import Path

from setuptools import setup, find_namespace_packages

__here__ = Path(__file__).parent.resolve()

sys.path.insert(0, str(__here__))

from autohooks.version import get_version
from autohooks.install import PostInstall, PostDevelop


with (__here__ / 'README.md').open('r') as f:
    long_description = f.read()

setup(
    name='autohooks',
    version=get_version(),
    author='Greenbone Networks GmbH',
    author_email='info@greenbone.net',
    description='Library for managing git hooks',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/bjoernricks/autohooks',
    packages=find_namespace_packages(include=['autohooks', 'autohooks.*']),
    package_data={'': ['template']},
    include_package_data=True,
    python_requires='>=3.5',
    classifiers=[
        # Full list: https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Version Control :: Git',
    ],
    install_requires=['toml'],
    cmdclass={"install": PostInstall, "develop": PostDevelop},
    entry_points={'console_scripts': ['autohooks=autohooks.cli:main']},
)



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from ..config import load_config_from_pyproject_toml, get_pyproject_toml_path
from ..install import (
    install_pre_commit_hook,
    get_pre_commit_hook_path,
    get_pre_commit_hook_template_path,
)


def install_hooks(args):
    pre_commit_hook = get_pre_commit_hook_path()
    pyproject_toml = get_pyproject_toml_path()
    config = load_config_from_pyproject_toml(pyproject_toml)

    if pre_commit_hook.exists() and not args.force:
        print(
            'pre-commit hook is already installed at {}. --force to '
            'override.'.format(str(pre_commit_hook))
        )
    else:
        if not config.is_autohooks_enabled():
            print(
                'Warning: autohooks is not enabled in your {} file. Please add '
                'a tools.autohooks section.'.format(str(pyproject_toml)),
                file=sys.stderr,
            )

        pre_commit_hook_template = get_pre_commit_hook_template_path()
        install_pre_commit_hook(pre_commit_hook_template, pre_commit_hook)

        print('pre-commit hook installed at {}'.format(str(pre_commit_hook)))



# Copyright (C) 2019 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shutil

from setuptools.command.install import install
from setuptools.command.develop import develop

from autohooks.utils import (
    get_git_hook_directory_path,
    get_autohooks_directory_path,
)


def get_pre_commit_hook_path():
    git_hook_dir_path = get_git_hook_directory_path()
    return git_hook_dir_path / 'pre-commit'


def get_pre_commit_hook_template_path():
    setup_dir_path = get_autohooks_directory_path()
    return setup_dir_path / 'precommit' / 'template'


def install_pre_commit_hook(pre_commit_hook_file, pre_commit_hook):
    shutil.copy(str(pre_commit_hook_file), str(pre_commit_hook))


class AutohooksInstall:
    def install_git_hook(self):
        try:
            pre_commit_hook = get_pre_commit_hook_path()
            if not pre_commit_hook.exists():
                pre_commit_hook_template = get_pre_commit_hook_template_path()
                install_pre_commit_hook(
                    pre_commit_hook_template, pre_commit_hook
                )
        except Exception:
            pass


class PostInstall(install, AutohooksInstall):
    def run(self):
        super().run()
        self.install_git_hook()


class PostDevelop(develop, AutohooksInstall):
    def install_for_development(self):
        super().install_for_development()
        self.install_git_hook()
